#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic service wrapper to access any API."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import datetime
import httplib
import sys
import time
import threading

from adspygoogle import SOAPpy
from adspygoogle.common import MessageHandler
from adspygoogle.common import SanityCheck
from adspygoogle.common import Utils
from adspygoogle.common.Errors import AuthTokenError
from adspygoogle.common.Errors import Error
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.Logger import Logger
from adspygoogle.SOAPpy.wstools.WSDLTools import WSDLError

sys_stdout_monkey_lock = threading.Lock()
# We will refresh an OAuth 2.0 credential _OAUTH2_REFRESH_MINUTES_IN_ADVANCE
# minutes in advance of it's expiration.
_OAUTH2_REFRESH_MINUTES_IN_ADVANCE = 5
# Reference SOAPpy.SOAPConfig object. Instances of GenericApiService will use
# a SOAPConfig copy constructor to have their own instance of SOAPConfig which
# starts off identical to this one.
_SOAP_CONFIG = SOAPpy.SOAPConfig(
    typed=0, namespaceStyle='2001', returnFaultInfo=1, dumpHeadersIn=1,
    dumpHeadersOut=1, dumpSOAPIn=1, dumpSOAPOut=1)


class GenericApiService(object):

  """Generic wrapper around a SOAPpy proxy.

  GenericApiService is an abstract class and should not be instantiated
  directly. Classes which implement this class are required to provide the
  following methods:

  _SetHeaders
  _GetMethodInfo
  _HandleLogsAndErrors

  The following methods are intended to be overridden as necessary, but provide
  functioning default implementations:

  _TakeActionOnSoapCall
  _TakeActionOnPackedArgs
  _WrapSoapCall
  """

  def __init__(self, headers, config, op_config, lock, logger, service_name,
               service_url, wrap_lists, buffer_class, namespace,
               namespace_extractor):
    """Inits GenericApiService.

    Args:
      headers: dict Dictionary object with populated authentication
               credentials.
      config: dict Dictionary object with populated configuration values.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.
      lock: mixed Thread lock to use to synchronize requests. May be a
            thread.lock or a threading.RLock
      logger: Logger Instance of Logger to use for logging.
      service_name: string The name of this service.
      service_url: string The URL pointing to this web service.
      wrap_lists: boolean Whether this service needs to wrap lists in an
                  additional layer of XML element tags.
      buffer_class: class The SoapBuffer subclass to use as a buffer.
      namespace: string The namespace this service uses by default.
      namespace_extractor: function A function which takes a URL and returns the
                           namespace prefix to use to represent it.

    Raises:
      Error: The WSDL for this service could not be found. Will also be raised
      if GenericApiService is initialized directly rather than by a subclass.
    """
    if self.__class__ == GenericApiService:
      raise Error('GenericApiService cannot be instantiated directly.')

    self._headers = headers
    self._config = config
    self._op_config = op_config
    self._lock = lock
    self._logger = logger
    self._service_name = service_name
    self._service_url = service_url
    self._wrap_lists = wrap_lists
    self._buffer_class = buffer_class
    self._namespace = namespace
    self._namespace_extractor = namespace_extractor
    self._method_proxies = {}

    wsdl_url = service_url + '?wsdl'
    try:
      self._soappyservice = SOAPpy.WSDL.Proxy(
          wsdl_url, noroot=1, http_proxy=self._op_config['http_proxy'],
          config=self._GetSoapConfig())
    except WSDLError:
      raise Error('Unable to locate WSDL at path \'%s\'' % wsdl_url)
    else:
      for method_key in self._soappyservice.methods:
        self._soappyservice.methods[method_key].location = service_url

  def _GetSoapConfig(self):
    """Creates a new SOAPpy.SOAPConfig for this service to use.

    This method is necessary to ensure that each SOAPConfig has a unique
    argsOrdering dictionary.

    Returns:
      SOAPpy.SOAPConfig A new SOAPConfig for this service to use.
    """
    config = SOAPpy.SOAPConfig(config=_SOAP_CONFIG)
    config.argsOrdering = {}
    return config

  def __getattr__(self, name):
    """Takes an attribute name and tries to create a SOAP call proxy around it.

    Args:
      name: string The name of the attribute to fetch. This should be the name
            of an operation/method in the SOAP service this object represents.

    Returns:
      function A function wrapping an invocation to the given operation
      (actually a function) in the SOAP service this object contains.

    Raises:
      AttributeError: if the given attribute name cannot be found in this
      object's SOAP service. Most likely, the given name was not a WSDL-defined
      SOAP operation in this service.
    """
    if name not in self._method_proxies:
      self._method_proxies[name] = self._WrapSoapCall(self._CreateMethod(name))
    return self._method_proxies[name]

  def __dir__(self):
    """Overrides default dir() behavior; prints the service's public methods."""
    dir_list = ['CallRawMethod']
    dir_list.extend(self._soappyservice.methods.keys())
    return dir_list

  def _WrapSoapCall(self, call_function):
    """Gives the service a chance to wrap a call in a product-specific function.

    If a product needs to wrap the call to a SOAP service, for example to listen
    for specific errors and retry failed calls, then its extending service class
    must override this method.

    Args:
      call_function: function The function to make a SOAP call.

    Returns:
      function The function to make a SOAP call. For some products, this will be
      the input function. For others, it may be a new function wrapping the
      input function.
    """
    return call_function

  def _SetHeaders(self):
    """Sets the SOAP headers for this service's requests.

    Must be overridden by an extending class.
    """
    raise NotImplementedError

  def _GetMethodInfo(self, method_name):
    """Pulls all of the relevant data about a method from a SOAPpy service.

    Must be overridden by an extending class.

    The return dictionary has two keys, MethodInfoKeys.INPUTS and
    MethodInfoKeys.OUTPUTS. Each of these keys has a list value. These lists
    contain a dictionary of information on the input/output parameter list, in
    order.

    Args:
      method_name: string The name of the method to pull information for.

    Returns:
      dict A dictionary containing information about a SOAP method.
    """
    raise NotImplementedError

  def _HandleLogsAndErrors(self, buf, start_time, stop_time, error=None):
    """Manage SOAP XML message.

    Must be overridden by an extending class.

    Args:
      buf: SoapBuffer SOAP buffer.
      start_time: str Time before service call was invoked.
      stop_time: str Time after service call was invoked.
      [optional]
      error: dict Error, if any.
    """
    raise NotImplementedError

  def _TakeActionOnSoapCall(self, method_name, args):
    """Gives the service a chance to take product-specific action on raw inputs.

    If a product needs to take the opportunity to modify the inputs, then its
    extending service class must override this method.

    Args:
      method_name: string The name of the SOAP operation being called.
      args: tuple The arguments passed into the SOAP operation.

    Returns:
      tuple The method arguments, possibly modified.
    """
    return args

  def _TakeActionOnPackedArgs(self, method_name, ksoap_args):
    """Allows a service to take product-specific action on packed arguments.

    If a product needs to take the opportunity to modify the packed inputs, then
    its extending service class must override this method.

    Args:
      method_name: string The name of the SOAP operation being called.
      ksoap_args: dictionary The keyword arguments packed for the SOAP
                  operation.

    Returns:
      dictionary The packed keyword arguments, possibly modified.
    """
    return ksoap_args

  def _ReadyOAuth(self):
    """If OAuth is on, sets the transport handler to add OAuth HTTP header.

    This method refreshes OAuth 2.0 credentials as necessary.
    """
    if self._headers.get('oauth2credentials'):
      self._RefreshCredentialIfNecessary(self._headers['oauth2credentials'])
      self._headers['oauth2credentials'].apply(
          self._soappyservice.soapproxy.transport.additional_headers)
    else:
      if ('Authorization' in
          self._soappyservice.soapproxy.transport.additional_headers):
        del self._soappyservice.soapproxy.transport.additional_headers[
            'Authorization']

  def _RefreshCredentialIfNecessary(self, credential):
    """Checks if the credential needs refreshing and refreshes if necessary."""
    if (credential.token_expiry is not None and credential.token_expiry -
        datetime.datetime.utcnow() <
        datetime.timedelta(minutes=_OAUTH2_REFRESH_MINUTES_IN_ADVANCE)):
      import httplib2
      self._headers['oauth2credentials'].refresh(httplib2.Http())

  def _ReadyCompression(self):
    """Sets whether the HTTP transport layer should use compression."""
    compress = Utils.BoolTypeConvert(self._config['compress'])
    self._soappyservice.soapproxy.config.send_compressed = compress
    self._soappyservice.soapproxy.config.accept_compressed = compress

  def _CreateMethod(self, method_name):
    """Create a method wrapping an invocation to the SOAP service."""
    try:
      soap_service_method = getattr(self._soappyservice, method_name)
    except AttributeError:
      method_name = method_name[0].lower() + method_name[1:]
      soap_service_method = getattr(self._soappyservice, method_name)

    def CallMethod(*args):
      """Perform a SOAP call."""
      try:
        self._lock.acquire()
        self._ReadyOAuth()
        self._ReadyCompression()
        self._SetHeaders()

        args = self._TakeActionOnSoapCall(method_name, args)
        method_info = self._GetMethodInfo(method_name)
        method_attrs_holder = None

        if len(method_info[MethodInfoKeys.INPUTS]) > 1:
          self._ConfigureArgOrder(method_name,
                                  method_info[MethodInfoKeys.INPUTS])

        if not method_info[MethodInfoKeys.INPUTS]:
          # Don't put any namespaces other than this service's namespace on
          # calls with no input params.
          method_attrs_holder = self._soappyservice.soapproxy.methodattrs
          self._soappyservice.soapproxy.methodattrs = {
              'xmlns': self._namespace
          }

        if len(args) != len(method_info[MethodInfoKeys.INPUTS]):
          raise TypeError(''.join([
              method_name + '() takes exactly ',
              str(len(self._soappyservice.methods[method_name].inparams)),
              ' argument(s). (', str(len(args)), ' given)']))

        ksoap_args = {}
        for i in range(len(method_info[MethodInfoKeys.INPUTS])):
          if Utils.BoolTypeConvert(self._config['strict']):
            SanityCheck.SoappySanityCheck(
                self._soappyservice, args[i],
                method_info[MethodInfoKeys.INPUTS][i][MethodInfoKeys.NS],
                method_info[MethodInfoKeys.INPUTS][i][MethodInfoKeys.TYPE],
                method_info[MethodInfoKeys.INPUTS][i][
                    MethodInfoKeys.MAX_OCCURS])

          element_name = str(method_info[MethodInfoKeys.INPUTS][i][
              MethodInfoKeys.ELEMENT_NAME])

          ksoap_args[element_name] = MessageHandler.PackForSoappy(
              args[i],
              method_info[MethodInfoKeys.INPUTS][i][MethodInfoKeys.NS],
              method_info[MethodInfoKeys.INPUTS][i][MethodInfoKeys.TYPE],
              self._soappyservice,
              self._wrap_lists,
              self._namespace_extractor)

        ksoap_args = self._TakeActionOnPackedArgs(method_name, ksoap_args)

        buf = self._buffer_class(
            xml_parser=self._config['xml_parser'],
            pretty_xml=Utils.BoolTypeConvert(self._config['pretty_xml']))
        sys_stdout_monkey_lock.acquire()
        try:
            old_stdout = sys.stdout
            sys.stdout = buf

            error = {}
            response = None
            start_time = time.strftime('%Y-%m-%d %H:%M:%S')
            try:
              response = MessageHandler.UnpackResponseAsDict(
                  soap_service_method(**ksoap_args))
            except Exception, e:
              error['data'] = e
            stop_time = time.strftime('%Y-%m-%d %H:%M:%S')
            # Restore stdout
            sys.stdout = old_stdout
        finally:
            sys_stdout_monkey_lock.release()

        if isinstance(response, Error):
          error = response

        if not Utils.BoolTypeConvert(self._config['raw_debug']):
          self._HandleLogsAndErrors(buf, start_time, stop_time, error)

        # When debugging mode is ON, fetch last traceback.
        if Utils.BoolTypeConvert(self._config['debug']):
          if Utils.LastStackTrace() and Utils.LastStackTrace() != 'None':
            error['trace'] = Utils.LastStackTrace()

        # Catch local errors prior to going down to the SOAP layer, which may
        # not exist for this error instance.
        if 'data' in error and not buf.IsHandshakeComplete():
          # Check if buffer contains non-XML data, most likely an HTML page.
          # This happens in the case of 502 errors (and similar). Otherwise,
          # this is a local error and API request was never made.
          html_error = Utils.GetErrorFromHtml(buf.GetBufferAsStr())
          if html_error:
            msg = html_error
          else:
            msg = str(error['data'])
            if Utils.BoolTypeConvert(self._config['debug']):
              msg += '\n%s' % error['trace']

          # When debugging mode is ON, store the raw content of the buffer.
          if Utils.BoolTypeConvert(self._config['debug']):
            error['raw_data'] = buf.GetBufferAsStr()

          # Catch errors from AuthToken and ValidationError levels, raised
          # during try/except above.
          if isinstance(error['data'], AuthTokenError):
            raise AuthTokenError(msg)
          elif isinstance(error['data'], ValidationError):
            raise ValidationError(error['data'])
          if 'raw_data' in error:
            msg = '%s [RAW DATA: %s]' % (msg, error['raw_data'])
          return Error(msg)

        if Utils.BoolTypeConvert(self._config['raw_response']):
          response = buf.GetRawSoapIn()
        elif error:
          response = error
        else:
          output_types = [(out_param[MethodInfoKeys.NS],
                           out_param[MethodInfoKeys.TYPE],
                           out_param[MethodInfoKeys.MAX_OCCURS]) for out_param
                          in method_info[MethodInfoKeys.OUTPUTS]]
          response = MessageHandler.RestoreListTypeWithSoappy(
              response, self._soappyservice, output_types)

        if Utils.BoolTypeConvert(self._config['wrap_in_tuple']):
          response = MessageHandler.WrapInTuple(response)

        # Restore method_attrs if they were over-ridden
        if method_attrs_holder:
          self._soappyservice.soapproxy.methodattrs = method_attrs_holder

        return response
      finally:
        self._lock.release()

    return CallMethod

  def _ConfigureArgOrder(self, method_name, inputs):
    """Ensure that SOAPpy knows what order in which to pack operation arguments.

    Even though we're pulling this information out of SOAPpy in the first place,
    it doesn't know to look there for argument ordering on its own...

    Args:
      method_name: str The name of the method to configure.
      inputs: list The input arguments for the given method.
    """
    if method_name not in self._soappyservice.soapproxy.config.argsOrdering:
      self._soappyservice.soapproxy.config.argsOrdering[method_name] = [
          argument[MethodInfoKeys.ELEMENT_NAME] for argument in inputs]

  def _ManageSoap(self, buf, log_handlers, lib_url, start_time, stop_time,
                  error=None):
    """Manage SOAP XML message.

    Args:
      buf: SoapBuffer SOAP buffer.
      log_handlers: list Log handlers.
      lib_url: str URL of the project's home.
      start_time: str Time before service call was invoked.
      stop_time: str Time after service call was invoked.
      [optional]
      error: dict Error, if any.

    Returns:
      mixed Any fault that occurred with this SOAP operation. The data type may
      be a string or dictionary depending on what went wrong. If there were no
      faults with this operation, None is returned.
    """
    if error is None:
      error = {}

    # Load trace errors, if any.
    if error and 'trace' in error:
      error_msg = error['trace']
    else:
      error_msg = ''

    # Check if response was successful or not.
    if error and 'data' in error:
      is_fault = True
    else:
      is_fault = False

    # Forward SOAP XML, errors, and other debugging data to console, external
    # file, both, or ignore. Each handler supports the following elements,
    #   tag: Config value for this handler. If left empty, will never write
    #        data to file.
    #   target: Target/destination represented by this handler (i.e. FILE,
    #           CONSOLE, etc.). Initially, it should be set to Logger.NONE.
    #   name: Name of the log file to use.
    #   data: Data to write.
    for handler in log_handlers:
      if handler['tag'] == 'xml_log':
        handler['target'] = Logger.NONE
        handler['data'] += ('StartTime: %s\n%s\n%s\n%s\n%s\nEndTime: %s'
                            % (start_time, buf.GetHeadersOut(),
                               buf.GetSoapOut(), buf.GetHeadersIn(),
                               buf.GetSoapIn(), stop_time))
      elif handler['tag'] == 'request_log':
        handler['target'] = Logger.NONE
        handler['data'] += ' isFault=%s' % is_fault
      elif not handler['tag']:
        handler['target'] = Logger.NONE
        handler['data'] += 'DEBUG: %s' % error_msg
    for handler in log_handlers:
      if (handler['tag'] and
          Utils.BoolTypeConvert(self._config[handler['tag']])):
        handler['target'] = Logger.FILE
      # If debugging is On, raise handler's target two levels,
      #   NONE -> CONSOLE
      #   FILE -> FILE_AND_CONSOLE.
      if Utils.BoolTypeConvert(self._config['debug']):
        handler['target'] += 2

      if (handler['target'] != Logger.NONE and handler['data'] and
          handler['data'] != 'None' and handler['data'] != 'DEBUG: '):
        self._logger.Log(handler['name'], handler['data'],
                         log_level=Logger.DEBUG, log_handler=handler['target'])

    # If raw response is requested, no need to validate and throw appropriate
    # error. Up to the end user to handle successful or failed request.
    if Utils.BoolTypeConvert(self._config['raw_response']): return

    # Report SOAP fault.
    if is_fault:
      try:
        fault = buf.GetFaultAsDict()
        if not fault: msg = error['data']
      except Exception:
        fault = None
        # An error is not a SOAP fault, but check if some other error.
        if error_msg:
          msg = error_msg
        else:
          msg = ('Unable to parse incoming SOAP XML. Please, file '
                 'a bug at %s/issues/list.' % lib_url)
      if not fault and msg: return msg
      return fault
    return None

  def CallRawMethod(self, soap_message):
    """Makes an API call by POSTing a raw SOAP XML message to the server.

    Args:
      soap_message: str SOAP XML message.

    Returns:
      tuple Raw XML response from the API method.

    Raises:
      Error: if the SOAP call is not successful. Most likely this is the result
      of the server sending back an HTTP error, such as a 502.
    """

    self._lock.acquire()
    try:
      buf = self._buffer_class(
          xml_parser=self._config['xml_parser'],
          pretty_xml=Utils.BoolTypeConvert(self._config['pretty_xml']))

      http_header = {
          'post': self._service_url,
          'host': Utils.GetNetLocFromUrl(self._op_config['server']),
          'user_agent': '%s; CallRawMethod' % self.__class__.__name__,
          'content_type': 'text/xml; charset=\"UTF-8\"',
          'content_length': '%d' % len(soap_message),
          'soap_action': ''
      }

      if self._headers.get('oauth2credentials'):
        self._headers['oauth2credentials'].apply(http_header)

      self._start_time = time.strftime('%Y-%m-%d %H:%M:%S')
      buf.write('%s Outgoing HTTP headers %s\nPOST %s\nHost: %s\nUser-Agent: '
                '%s\nContent-type: %s\nContent-length: %s\nSOAPAction: %s\n' %
                ('*'*3, '*'*46, http_header['post'], http_header['host'],
                 http_header['user_agent'], http_header['content_type'],
                 http_header['content_length'], http_header['soap_action']))
      if self._headers.get('oauth2credentials'):
        buf.write('Authorization: ' + http_header['Authorization'] + '\n')
      buf.write('%s\n%s Outgoing SOAP %s\n%s\n%s\n' %
                ('*'*72, '*'*3, '*'*54, soap_message, '*'*72))

      if self._op_config['http_proxy']:
        real_address = self._op_config['http_proxy']
      else:
        real_address = http_header['host']

      # Construct header and send SOAP message.
      web_service = httplib.HTTPS(real_address)
      web_service.putrequest('POST', http_header['post'])
      web_service.putheader('Host', http_header['host'])
      web_service.putheader('User-Agent', http_header['user_agent'])
      web_service.putheader('Content-type', http_header['content_type'])
      web_service.putheader('Content-length', http_header['content_length'])
      web_service.putheader('SOAPAction', http_header['soap_action'])
      if self._headers.get('oauth2credentials'):
        web_service.putheader('Authorization', http_header['Authorization'])
      web_service.endheaders()
      web_service.send(soap_message)

      # Get response.
      status_code, status_message, header = web_service.getreply()
      response = web_service.getfile().read()

      header = str(header).replace('\r', '')
      buf.write(('%s Incoming HTTP headers %s\n%s %s\n%s\n%s\n%s Incoming SOAP'
                 ' %s\n%s\n%s\n' % ('*'*3, '*'*46, status_code, status_message,
                                    header, '*'*72, '*'*3, '*'*54, response,
                                    '*'*72)))
      self._stop_time = time.strftime('%Y-%m-%d %H:%M:%S')

      # Catch local errors prior to going down to the SOAP layer, which may not
      # exist for this error instance.
      if not buf.IsHandshakeComplete() or not buf.IsSoap():
        # The buffer contains non-XML data, most likely an HTML page. This
        # happens in the case of 502 errors.
        html_error = Utils.GetErrorFromHtml(buf.GetBufferAsStr())
        if html_error:
          msg = html_error
        else:
          msg = 'Unknown error.'
        raise Error(msg)

      self._HandleLogsAndErrors(buf, self._start_time, self._stop_time)
    finally:
      self._lock.release()
    if self._config['wrap_in_tuple']:
      response = MessageHandler.WrapInTuple(response)
    return response


class MethodInfoKeys(object):
  """Static constants holder; keys used to pass method information around."""

  INPUTS = 'inputs'
  ELEMENT_NAME = 'elemname'
  NS = 'ns'
  TYPE = 'type'
  OUTPUTS = 'outputs'
  MAX_OCCURS = 'maxOccurs'



#! /usr/bin/env python
"""Namespace module, so you don't need PyXML 
"""

try:
    from xml.ns import SOAP, SCHEMA, WSDL, XMLNS, DSIG, ENCRYPTION
    DSIG.C14N       = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
    
except:
    class SOAP:
        ENV         = "http://schemas.xmlsoap.org/soap/envelope/"
        ENC         = "http://schemas.xmlsoap.org/soap/encoding/"
        ACTOR_NEXT  = "http://schemas.xmlsoap.org/soap/actor/next"

    class SCHEMA:
        XSD1        = "http://www.w3.org/1999/XMLSchema"
        XSD2        = "http://www.w3.org/2000/10/XMLSchema"
        XSD3        = "http://www.w3.org/2001/XMLSchema"
        XSD_LIST    = [ XSD1, XSD2, XSD3 ]
        XSI1        = "http://www.w3.org/1999/XMLSchema-instance"
        XSI2        = "http://www.w3.org/2000/10/XMLSchema-instance"
        XSI3        = "http://www.w3.org/2001/XMLSchema-instance"
        XSI_LIST    = [ XSI1, XSI2, XSI3 ]
        BASE        = XSD3

    class WSDL:
        BASE        = "http://schemas.xmlsoap.org/wsdl/"
        BIND_HTTP   = "http://schemas.xmlsoap.org/wsdl/http/"
        BIND_MIME   = "http://schemas.xmlsoap.org/wsdl/mime/"
        BIND_SOAP   = "http://schemas.xmlsoap.org/wsdl/soap/"
        BIND_SOAP12 = "http://schemas.xmlsoap.org/wsdl/soap12/"

    class XMLNS:
        BASE        = "http://www.w3.org/2000/xmlns/"
        XML         = "http://www.w3.org/XML/1998/namespace"
        HTML        = "http://www.w3.org/TR/REC-html40"

    class DSIG:
        BASE         = "http://www.w3.org/2000/09/xmldsig#"
        C14N         = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
        C14N_COMM    = "http://www.w3.org/TR/2000/CR-xml-c14n-20010315#WithComments"
        C14N_EXCL    = "http://www.w3.org/2001/10/xml-exc-c14n#"
        DIGEST_MD2   = "http://www.w3.org/2000/09/xmldsig#md2"
        DIGEST_MD5   = "http://www.w3.org/2000/09/xmldsig#md5"
        DIGEST_SHA1  = "http://www.w3.org/2000/09/xmldsig#sha1"
        ENC_BASE64   = "http://www.w3.org/2000/09/xmldsig#base64"
        ENVELOPED    = "http://www.w3.org/2000/09/xmldsig#enveloped-signature"
        HMAC_SHA1    = "http://www.w3.org/2000/09/xmldsig#hmac-sha1"
        SIG_DSA_SHA1 = "http://www.w3.org/2000/09/xmldsig#dsa-sha1"
        SIG_RSA_SHA1 = "http://www.w3.org/2000/09/xmldsig#rsa-sha1"
        XPATH        = "http://www.w3.org/TR/1999/REC-xpath-19991116"
        XSLT         = "http://www.w3.org/TR/1999/REC-xslt-19991116"

    class ENCRYPTION:
        BASE    = "http://www.w3.org/2001/04/xmlenc#"
        BLOCK_3DES    = "http://www.w3.org/2001/04/xmlenc#des-cbc"
        BLOCK_AES128    = "http://www.w3.org/2001/04/xmlenc#aes128-cbc"
        BLOCK_AES192    = "http://www.w3.org/2001/04/xmlenc#aes192-cbc"
        BLOCK_AES256    = "http://www.w3.org/2001/04/xmlenc#aes256-cbc"
        DIGEST_RIPEMD160    = "http://www.w3.org/2001/04/xmlenc#ripemd160"
        DIGEST_SHA256    = "http://www.w3.org/2001/04/xmlenc#sha256"
        DIGEST_SHA512    = "http://www.w3.org/2001/04/xmlenc#sha512"
        KA_DH    = "http://www.w3.org/2001/04/xmlenc#dh"
        KT_RSA_1_5    = "http://www.w3.org/2001/04/xmlenc#rsa-1_5"
        KT_RSA_OAEP    = "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p"
        STREAM_ARCFOUR    = "http://www.w3.org/2001/04/xmlenc#arcfour"
        WRAP_3DES    = "http://www.w3.org/2001/04/xmlenc#kw-3des"
        WRAP_AES128    = "http://www.w3.org/2001/04/xmlenc#kw-aes128"
        WRAP_AES192    = "http://www.w3.org/2001/04/xmlenc#kw-aes192"
        WRAP_AES256    = "http://www.w3.org/2001/04/xmlenc#kw-aes256"


class OASIS:
    '''URLs for Oasis specifications
    '''
    WSSE    = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"
    UTILITY = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
    LIFETIME = "http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd"
    PROPERTIES = "http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.xsd"
    BASENOTIFICATION = "http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd"
    BASEFAULTS = "http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-BaseFaults-1.2-draft-01.xsd"

class WSSE:
    BASE    = "http://schemas.xmlsoap.org/ws/2002/04/secext"
    TRUST   = "http://schemas.xmlsoap.org/ws/2004/04/trust"


class WSU:
    BASE    = "http://schemas.xmlsoap.org/ws/2002/04/utility"
    UTILITY = "http://schemas.xmlsoap.org/ws/2002/07/utility"


class WSR:
    PROPERTIES = "http://www.ibm.com/xmlns/stdwip/web-services/WS-ResourceProperties"
    LIFETIME   = "http://www.ibm.com/xmlns/stdwip/web-services/WS-ResourceLifetime"


class WSA200408:
    ADDRESS    = "http://schemas.xmlsoap.org/ws/2004/08/addressing"
    ANONYMOUS  = "%s/role/anonymous" %ADDRESS
    FAULT      = "%s/fault" %ADDRESS
WSA = WSA200408

class WSA200403:
    ADDRESS    = "http://schemas.xmlsoap.org/ws/2004/03/addressing"
    ANONYMOUS  = "%s/role/anonymous" %ADDRESS
    FAULT      = "%s/fault" %ADDRESS

class WSA200303:
    ADDRESS    = "http://schemas.xmlsoap.org/ws/2003/03/addressing"
    ANONYMOUS  = "%s/role/anonymous" %ADDRESS
    FAULT      = None

class WSP:
    POLICY = "http://schemas.xmlsoap.org/ws/2002/12/policy"

class BEA:
    SECCONV = "http://schemas.xmlsoap.org/ws/2004/04/sc"

class GLOBUS:
    SECCONV = "http://wsrf.globus.org/core/2004/07/security/secconv"
    CORE    = "http://www.globus.org/namespaces/2004/06/core"
    SIG     = "http://www.globus.org/2002/04/xmlenc#gssapi-sign"

class JAXB:
    NS = "http://java.sun.com/xml/ns/jaxb"

ZSI_SCHEMA_URI = 'http://www.zolera.com/schemas/ZSI/'



#! /usr/bin/env python
"""WSDL parsing services package for Web Services for Python."""

ident = "$Id: __init__.py,v 1.11 2004/12/07 15:54:53 blunck2 Exp $"

import WSDLTools
import XMLname
import logging




"""
################################################################################
# Copyright (c) 2003, Pfizer
# Copyright (c) 2001, Cayce Ullman.
# Copyright (c) 2001, Brian Matthews.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of actzero, inc. nor the names of its contributors may
# be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################
"""

ident = '$Id: Utilities.py,v 1.4 2004/01/31 04:20:06 warnes Exp $'
from version import __version__

import exceptions
import copy
import re
import string
import sys
from types import *

# SOAPpy modules
from Errors import *

################################################################################
# Utility infielders
################################################################################
def collapseWhiteSpace(s):
    return re.sub('\s+', ' ', s).strip()

def decodeHexString(data):
    conv = {
            '0': 0x0, '1': 0x1, '2': 0x2, '3': 0x3, '4': 0x4,
            '5': 0x5, '6': 0x6, '7': 0x7, '8': 0x8, '9': 0x9,
            
            'a': 0xa, 'b': 0xb, 'c': 0xc, 'd': 0xd, 'e': 0xe,
            'f': 0xf,
            
            'A': 0xa, 'B': 0xb, 'C': 0xc, 'D': 0xd, 'E': 0xe,
            'F': 0xf,
            }
    
    ws = string.whitespace

    bin = ''

    i = 0

    while i < len(data):
        if data[i] not in ws:
            break
        i += 1

    low = 0

    while i < len(data):
        c = data[i]

        if c in string.whitespace:
            break

        try:
            c = conv[c]
        except KeyError:
            raise ValueError, \
                "invalid hex string character `%s'" % c

        if low:
            bin += chr(high * 16 + c)
            low = 0
        else:
            high = c
            low = 1

        i += 1

    if low:
        raise ValueError, "invalid hex string length"

    while i < len(data):
        if data[i] not in string.whitespace:
            raise ValueError, \
                "invalid hex string character `%s'" % c

        i += 1

    return bin

def encodeHexString(data):
    h = ''

    for i in data:
        h += "%02X" % ord(i)

    return h

def leapMonth(year, month):
    return month == 2 and \
        year % 4 == 0 and \
        (year % 100 != 0 or year % 400 == 0)

def cleanDate(d, first = 0):
    ranges = (None, (1, 12), (1, 31), (0, 23), (0, 59), (0, 61))
    months = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
    names = ('year', 'month', 'day', 'hours', 'minutes', 'seconds')

    if len(d) != 6:
        raise ValueError, "date must have 6 elements"

    for i in range(first, 6):
        s = d[i]

        if type(s) == FloatType:
            if i < 5:
                try:
                    s = int(s)
                except OverflowError:
                    if i > 0:
                        raise
                    s = long(s)

                if s != d[i]:
                    raise ValueError, "%s must be integral" % names[i]

                d[i] = s
        elif type(s) == LongType:
            try: s = int(s)
            except: pass
        elif type(s) != IntType:
            raise TypeError, "%s isn't a valid type" % names[i]

        if i == first and s < 0:
            continue

        if ranges[i] != None and \
            (s < ranges[i][0] or ranges[i][1] < s):
            raise ValueError, "%s out of range" % names[i]

    if first < 6 and d[5] >= 61:
        raise ValueError, "seconds out of range"

    if first < 2:
        leap = first < 1 and leapMonth(d[0], d[1])

        if d[2] > months[d[1]] + leap:
            raise ValueError, "day out of range"

def debugHeader(title):
    s = '*** ' + title + ' '
    print s + ('*' * (72 - len(s)))

def debugFooter(title):
    print '*' * 72
    sys.stdout.flush()



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic proxy to access any DFA web service."""

__author__ = 'api.jdilallo@gmail.com (Joseph DiLallo)'

import warnings

from adspygoogle import SOAPpy
from adspygoogle.common import Utils
from adspygoogle.common.Errors import Error
from adspygoogle.common.GenericApiService import GenericApiService
from adspygoogle.common.GenericApiService import MethodInfoKeys
from adspygoogle.dfa import DfaUtils
from adspygoogle.dfa import LIB_SIG
from adspygoogle.dfa import LIB_URL
from adspygoogle.dfa import WSSE_NS
from adspygoogle.dfa.DfaErrors import DfaApiError
from adspygoogle.dfa.DfaErrors import DfaAuthenticationError
from adspygoogle.dfa.DfaSoapBuffer import DfaSoapBuffer


_DEPRECATION_WARNING = ('Legacy DFA passwords are deprecated. Please use '
                        'OAuth 2.0')
warnings.filterwarnings('always', _DEPRECATION_WARNING, DeprecationWarning)


class GenericDfaService(GenericApiService):

  """Wrapper for any DFA web service."""

  # The _WRAP_LISTS constant indicates that all DFA services need to wrap lists
  # in an extra layer of XML element tags.
  _WRAP_LISTS = True
  # The _BUFFER_CLASS is the subclass of SoapBuffer that should be used to track
  # all SOAP interactions.
  _BUFFER_CLASS = DfaSoapBuffer
  # The _TOKEN_EXPIRED_ERROR_MESSAGE is returned by the DFA API when a DFA token
  # needs to be refreshed.
  _TOKEN_EXPIRED_ERROR_MESSAGE = 'Authentication token has expired.'

  def __init__(self, headers, config, op_config, lock, logger, service_name):
    """Inits GenericDfaService.

    Args:
      headers: dict Dictionary object with populated authentication
               credentials.
      config: dict Dictionary object with populated configuration values.
      op_config: dict Dictionary object with additional configuration values for
                 this operation.
      lock: threading.RLock Thread lock to use to synchronize requests.
      logger: Logger Instance of Logger to use for logging.
      service_name: string The name of this service.
    """
    service_url = '/'.join([op_config['server'], op_config['version'],
                            'api/dfa-api', service_name])
    namespace = '/'.join(['http://www.doubleclick.net/dfa-api',
                          op_config['version']])
    namespace_extractor = _DetermineNamespacePrefix

    super(GenericDfaService, self).__init__(
        headers, config, op_config, lock, logger, service_name, service_url,
        GenericDfaService._WRAP_LISTS, GenericDfaService._BUFFER_CLASS,
        namespace, namespace_extractor)

    # DFA-specific changes to the SOAPpy.WSDL.Proxy
    methodattrs = {
        'xmlns:dfa': self._namespace
    }
    self._soappyservice.soapproxy.methodattrs = methodattrs

  def _WrapSoapCall(self, soap_call_function):
    """Gives the service a chance to wrap a call in a product-specific function.

    DFA uses this function to listen for expired DDMM tokens and refresh them.
    Calls which fail due to expired tokens will be retried.

    Args:
      soap_call_function: function The function to make a SOAP call.

    Returns:
      function A new function wrapping the input function which listens for
      token expired errors and retries the failed call.
    """

    def RefreshTokenIfExpired(*args, **kargs):
      try:
        return soap_call_function(*args, **kargs)
      except DfaAuthenticationError, e:
        if e.message == self._TOKEN_EXPIRED_ERROR_MESSAGE:
          self._GenerateToken()
          return soap_call_function(*args, **kargs)
        else:
          raise e

    return RefreshTokenIfExpired

  def _SetHeaders(self):
    """Sets the SOAP headers for this service's requests."""
    soap_headers = SOAPpy.Types.headerType()
    if self._service_name != 'login':
      if 'AuthToken' not in self._headers or not self._headers['AuthToken']:
        self._GenerateToken()
      wsse_header = SOAPpy.Types.structType(
          data={
              'UsernameToken': {
                  'Username': self._headers['Username'],
                  'Password': self._headers['AuthToken']
              }
          },
          name='Security', typed=0, attrs={'xmlns': WSSE_NS})
      soap_headers.Security = wsse_header
    request_header = SOAPpy.Types.structType(
        data={'applicationName': ''.join([self._headers['appName'], LIB_SIG])},
        name='RequestHeader', typed=0)
    soap_headers.RequestHeader = request_header
    self._soappyservice.soapproxy.header = soap_headers

  def _ReadyOAuth(self):
    """If OAuth is on, sets the transport handler to add OAuth2 HTTP header.

    DFA overrides the default implementation because only the login service
    should have this header.
    """
    if self._service_name == 'login':
      super(GenericDfaService, self)._ReadyOAuth()

  def _GetMethodInfo(self, method_name):
    """Pulls all of the relevant data about a method from a SOAPpy service.

    The return dictionary has two keys, MethodInfoKeys.INPUTS and
    MethodInfoKeys.OUTPUTS. Each of these keys has a list value. The list
    value contains a dictionary of information on each input/output parameter,
    in order.

    Args:
      method_name: string The name of the method to pull information for.

    Returns:
      dict A dictionary containing information about a SOAP method.
    """
    rval = {}
    rval[MethodInfoKeys.INPUTS] = []
    for i in range(len(self._soappyservice.methods[method_name].inparams)):
      param_attributes = self._soappyservice.methods[method_name].inparams[i]
      if hasattr(param_attributes, 'maxOccurs'):
        max_occurs = param_attributes.maxOccurs
      else:
        max_occurs = '1'
      inparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes.name,
          MethodInfoKeys.NS: param_attributes.type[0],
          MethodInfoKeys.TYPE: param_attributes.type[1],
          MethodInfoKeys.MAX_OCCURS: max_occurs
      }
      rval[MethodInfoKeys.INPUTS].append(inparam)

    rval[MethodInfoKeys.OUTPUTS] = []
    for i in range(len(self._soappyservice.methods[method_name].outparams)):
      param_attributes = self._soappyservice.methods[method_name].outparams[i]
      if hasattr(param_attributes, 'maxOccurs'):
        max_occurs = param_attributes.maxOccurs
      else:
        max_occurs = '1'
      outparam = {
          MethodInfoKeys.ELEMENT_NAME: param_attributes.name,
          MethodInfoKeys.NS: param_attributes.type[0],
          MethodInfoKeys.TYPE: param_attributes.type[1],
          MethodInfoKeys.MAX_OCCURS: max_occurs
      }
      rval[MethodInfoKeys.OUTPUTS].append(outparam)
    return rval

  def _TakeActionOnSoapCall(self, method_name, args):
    """Gives the service a chance to take product-specific action on raw inputs.

    DFA will try to determine xsi_types for saveAd and saveCreative calls.

    Args:
      method_name: string The name of the SOAP operation being called.
      args: tuple The arguments passed into the SOAP operation.

    Returns:
      tuple The method arguments, possibly modified.
    """
    if method_name.lower() == 'savecreative':
      DfaUtils.AssignCreativeXsi(args[0])
    elif method_name.lower() == 'savead':
      DfaUtils.AssignAdXsi(args[0])
    return args

  def _ReadyCompression(self):
    """Sets whether the HTTP transport layer should use compression.

    Overloaded for DFA because the DFA servers do not accept compressed
    messages. They do support returning compressed messages.
    """
    compress = Utils.BoolTypeConvert(self._config['compress'])
    self._soappyservice.soapproxy.config.send_compressed = False
    self._soappyservice.soapproxy.config.accept_compressed = compress

  def _HandleLogsAndErrors(self, buf, start_time, stop_time, error=None):
    """Manage SOAP XML message.

    Args:
      buf: SoapBuffer SOAP buffer.
      start_time: str Time before service call was invoked.
      stop_time: str Time after service call was invoked.
      [optional]
      error: dict Error, if any.

    Raises:
      DfaApiError: if the API calls returns a SOAP error message.
      Error: if the call returns a non-SOAP error message, such as an HTTP 502.
    """
    if error is None:
      error = {}
    try:
      handlers = self.__GetLogHandlers(buf)

      fault = super(GenericDfaService, self)._ManageSoap(
          buf, handlers, LIB_URL, start_time, stop_time, error)
      if fault:
        # Raise a specific error, subclass of DfaApiError.
        if fault['detail'] is None: del fault['detail']
        if 'detail' in fault:
          if ('google' in fault['detail'] and
              'doubleclick' not in fault['detail']):
            fault['detail']['doubleclick'] = fault['detail']['google']
          if ('doubleclick' in fault['detail'] and
              'errorCode' in fault['detail']['doubleclick']):
            code = int(fault['detail']['doubleclick']['errorCode'])
            if code == 4:
              raise DfaAuthenticationError(fault)
            else:
              raise DfaApiError(fault)
        if isinstance(fault, (str, dict)):
          raise DfaApiError(fault)
    except DfaApiError, e:
      raise e
    except Error, e:
      if error: e = error
      raise Error(e)

  def _GenerateToken(self):
    """Attempts to generate a token for the WSSE security header.

    Raises:
      DfaAuthenticationError: if there are not enough credentials to generate a
                              token or if the given credentials are invalid.
    """
    if ('Username' in self._headers and
        ('Password' in self._headers or 'oauth2credentials' in self._headers)):
      if not self._headers.get('oauth2credentials'):
        warnings.warn(_DEPRECATION_WARNING, DeprecationWarning, stacklevel=5)
      # Ensure the 'raw_response' config value is off while generating tokens.
      old_raw_response = self._config['raw_response']
      self._config['raw_response'] = 'n'
      try:
        login_service = GenericDfaService(
            self._headers, self._config, self._op_config, self._lock,
            self._logger, 'login')
        self._headers['AuthToken'] = login_service.authenticate(
            self._headers['Username'],
            self._headers.get('Password'))[0]['token']
      finally:
        self._config['raw_response'] = old_raw_response
    else:
      fault = {
          'faultstring': ('Authentication data, username/password or username/'
                          'oauth2credentials, is missing.')
      }
      raise DfaAuthenticationError(fault)

  def __GetLogHandlers(self, buf):
    """Gets a list of log handlers for the DFA library.

    Args:
      buf: SoapBuffer SOAP buffer from which calls are retrieved for logging.

    Returns:
      list Log handlers for the DFA library.
    """
    return [
        {
            'tag': 'xml_log',
            'name': 'soap_xml',
            'data': ''
        },
        {
            'tag': 'request_log',
            'name': 'request_info',
            'data': str('host=%s service=%s method=%s responseTime=%s '
                        'requestID=%s'
                        % (Utils.GetNetLocFromUrl(self._service_url),
                           self._service_name, buf.GetCallName(),
                           buf.GetCallResponseTime(), buf.GetCallRequestId()))
        },
        {
            'tag': '',
            'name': 'dfa_api_lib',
            'data': ''
        }
    ]


def _DetermineNamespacePrefix(unused_url):
  """Returns the SOAP prefix to use for definitions within the given namespace.

  Args:
    unused_url: string The URL of the namespace. The DFA library doesn't
                actually check this value.

  Returns:
    string The SOAP namespace prefix to use for the given namespace. The DFA
           library always returns 'dfa:'.
  """
  return 'dfa:'



# SOAPpy modules
from Config    import Config
from Types     import *
from NS        import NS
from Utilities import *

import string
import fpconst
import xml.sax
from wstools.XMLname import fromXMLname

ident = '$Id: Parser.py,v 1.16 2005/02/22 04:29:42 warnes Exp $'
from version import __version__


################################################################################
# SOAP Parser
################################################################################
class RefHolder:
    def __init__(self, name, frame):
        self.name = name
        self.parent = frame
        self.pos = len(frame)
        self.subpos = frame.namecounts.get(name, 0)

    def __repr__(self):
        return "<%s %s at %d>" % (self.__class__, self.name, id(self))

    def __str__(self):
        return "<%s %s at %d>" % (self.__class__, self.name, id(self))

class SOAPParser(xml.sax.handler.ContentHandler):
    class Frame:
        def __init__(self, name, kind = None, attrs = {}, rules = {}):
            self.name = name
            self.kind = kind
            self.attrs = attrs
            self.rules = rules

            self.contents = []
            self.names = []
            self.namecounts = {}
            self.subattrs = []

        def append(self, name, data, attrs):
            self.names.append(name)
            self.contents.append(data)
            self.subattrs.append(attrs)

            if self.namecounts.has_key(name):
                self.namecounts[name] += 1
            else:
                self.namecounts[name] = 1

        def _placeItem(self, name, value, pos, subpos = 0, attrs = None):
            self.contents[pos] = value

            if attrs:
                self.attrs.update(attrs)

        def __len__(self):
            return len(self.contents)

        def __repr__(self):
            return "<%s %s at %d>" % (self.__class__, self.name, id(self))

    def __init__(self, rules = None):
        xml.sax.handler.ContentHandler.__init__(self)
        self.body       = None
        self.header     = None
        self.attrs      = {}
        self._data      = None
        self._next      = "E" # Keeping state for message validity
        self._stack     = [self.Frame('SOAP')]

        # Make two dictionaries to store the prefix <-> URI mappings, and
        # initialize them with the default
        self._prem      = {NS.XML_T: NS.XML}
        self._prem_r    = {NS.XML: NS.XML_T}
        self._ids       = {}
        self._refs      = {}
        self._rules    = rules

    def startElementNS(self, name, qname, attrs):
        # Workaround two sax bugs
        if name[0] == None and name[1][0] == ' ':
            name = (None, name[1][1:])
        else:
            name = tuple(name)

        # First some checking of the layout of the message

        if self._next == "E":
            if name[1] != 'Envelope':
                raise Error, "expected `SOAP-ENV:Envelope', gto `%s:%s'" % \
                    (self._prem_r[name[0]], name[1])
            if name[0] != NS.ENV:
                raise faultType, ("%s:VersionMismatch" % NS.ENV_T,
                    "Don't understand version `%s' Envelope" % name[0])
            else:
                self._next = "HorB"
        elif self._next == "HorB":
            if name[0] == NS.ENV and name[1] in ("Header", "Body"):
                self._next = None
            else:
                raise Error, \
                    "expected `SOAP-ENV:Header' or `SOAP-ENV:Body', " \
                    "got `%s'" % self._prem_r[name[0]] + ':' + name[1]
        elif self._next == "B":
            if name == (NS.ENV, "Body"):
                self._next = None
            else:
                raise Error, "expected `SOAP-ENV:Body', got `%s'" % \
                    self._prem_r[name[0]] + ':' + name[1]
        elif self._next == "":
            raise Error, "expected nothing, got `%s'" % \
                self._prem_r[name[0]] + ':' + name[1]

        if len(self._stack) == 2:
            rules = self._rules
        else:
            try:
                rules = self._stack[-1].rules[name[1]]
            except:
                rules = None

        if type(rules) not in (NoneType, DictType):
            kind = rules
        else:
            kind = attrs.get((NS.ENC, 'arrayType'))

            if kind != None:
                del attrs._attrs[(NS.ENC, 'arrayType')]

                i = kind.find(':')
                if i >= 0:
                    kind = (self._prem[kind[:i]], kind[i + 1:])
                else:
                    kind = None

        self.pushFrame(self.Frame(name[1], kind, attrs._attrs, rules))

        self._data = [] # Start accumulating

    def pushFrame(self, frame):
        self._stack.append(frame)

    def popFrame(self):
        return self._stack.pop()

    def endElementNS(self, name, qname):
        # Workaround two sax bugs
        if name[0] == None and name[1][0] == ' ':
            ns, name = None, name[1][1:]
        else:
            ns, name = tuple(name)

        name = fromXMLname(name) # convert to SOAP 1.2 XML name encoding

        if self._next == "E":
            raise Error, "didn't get SOAP-ENV:Envelope"
        if self._next in ("HorB", "B"):
            raise Error, "didn't get SOAP-ENV:Body"

        cur = self.popFrame()
        attrs = cur.attrs

        idval = None

        if attrs.has_key((None, 'id')):
            idval = attrs[(None, 'id')]

            if self._ids.has_key(idval):
                raise Error, "duplicate id `%s'" % idval

            del attrs[(None, 'id')]

        root = 1

        if len(self._stack) == 3:
            if attrs.has_key((NS.ENC, 'root')):
                root = int(attrs[(NS.ENC, 'root')])

                # Do some preliminary checks. First, if root="0" is present,
                # the element must have an id. Next, if root="n" is present,
                # n something other than 0 or 1, raise an exception.

                if root == 0:
                    if idval == None:
                        raise Error, "non-root element must have an id"
                elif root != 1:
                    raise Error, "SOAP-ENC:root must be `0' or `1'"

                del attrs[(NS.ENC, 'root')]

        while 1:
            href = attrs.get((None, 'href'))
            if href:
                if href[0] != '#':
                    raise Error, "Non-local hrefs are not yet suppported."
                if self._data != None and \
                   string.join(self._data, "").strip() != '':
                    raise Error, "hrefs can't have data"

                href = href[1:]

                if self._ids.has_key(href):
                    data = self._ids[href]
                else:
                    data = RefHolder(name, self._stack[-1])

                    if self._refs.has_key(href):
                        self._refs[href].append(data)
                    else:
                        self._refs[href] = [data]

                del attrs[(None, 'href')]

                break

            kind = None

            if attrs:
                for i in NS.XSI_L:
                    if attrs.has_key((i, 'type')):
                        kind = attrs[(i, 'type')]
                        del attrs[(i, 'type')]

                if kind != None:
                    i = kind.find(':')
                    if i >= 0:
                        kind = (self._prem[kind[:i]], kind[i + 1:])
                    else:
# XXX What to do here? (None, kind) is just going to fail in convertType
                        #print "Kind with no NS:", kind
                        kind = (None, kind)

            null = 0

            if attrs:
                for i in (NS.XSI, NS.XSI2):
                    if attrs.has_key((i, 'null')):
                        null = attrs[(i, 'null')]
                        del attrs[(i, 'null')]

                if attrs.has_key((NS.XSI3, 'nil')):
                    null = attrs[(NS.XSI3, 'nil')]
                    del attrs[(NS.XSI3, 'nil')]


                ## Check for nil

                # check for nil='true'
                if type(null) in (StringType, UnicodeType):
                    if null.lower() == 'true':
                        null = 1

                # check for nil=1, but watch out for string values
                try:                
                    null = int(null)
                except ValueError, e:
                    if not e[0].startswith("invalid literal for int()"):
                        raise e
                    null = 0

                if null:
                    if len(cur) or \
                        (self._data != None and string.join(self._data, "").strip() != ''):
                        raise Error, "nils can't have data"

                    data = None

                    break

            if len(self._stack) == 2:
                if (ns, name) == (NS.ENV, "Header"):
                    self.header = data = headerType(attrs = attrs)
                    self._next = "B"
                    break
                elif (ns, name) == (NS.ENV, "Body"):
                    self.body = data = bodyType(attrs = attrs)
                    self._next = ""
                    break
            elif len(self._stack) == 3 and self._next == None:
                if (ns, name) == (NS.ENV, "Fault"):
                    data = faultType()
                    self._next = None # allow followons
                    break

            #print "\n"
            #print "data=", self._data
            #print "kind=", kind
            #print "cur.kind=", cur.kind
            #print "cur.rules=", cur.rules
            #print "\n"
                        

            if cur.rules != None:
                rule = cur.rules

                if type(rule) in (StringType, UnicodeType):
                    rule = (None, rule) # none flags special handling
                elif type(rule) == ListType:
                    rule = tuple(rule)

                #print "kind=",kind
                #print "rule=",rule


# XXX What if rule != kind?
                if callable(rule):
                    data = rule(string.join(self._data, ""))
                elif type(rule) == DictType:
                    data = structType(name = (ns, name), attrs = attrs)
                elif rule[1][:9] == 'arrayType':
                    data = self.convertType(cur.contents,
                                            rule, attrs)
                else:
                    data = self.convertType(string.join(self._data, ""),
                                            rule, attrs)

                break

            #print "No rules, using kind or cur.kind..."

            if (kind == None and cur.kind != None) or \
                (kind == (NS.ENC, 'Array')):
                kind = cur.kind

                if kind == None:
                    kind = 'ur-type[%d]' % len(cur)
                else:
                    kind = kind[1]

                if len(cur.namecounts) == 1:
                    elemsname = cur.names[0]
                else:
                    elemsname = None

                data = self.startArray((ns, name), kind, attrs, elemsname)

                break

            if len(self._stack) == 3 and kind == None and \
                len(cur) == 0 and \
                (self._data == None or string.join(self._data, "").strip() == ''):
                data = structType(name = (ns, name), attrs = attrs)
                break

            if len(cur) == 0 and ns != NS.URN:
                # Nothing's been added to the current frame so it must be a
                # simple type.

                if kind == None:
                    # If the current item's container is an array, it will
                    # have a kind. If so, get the bit before the first [,
                    # which is the type of the array, therefore the type of
                    # the current item.

                    kind = self._stack[-1].kind

                    if kind != None:
                        i = kind[1].find('[')
                        if i >= 0:
                            kind = (kind[0], kind[1][:i])
                    elif ns != None:
                        kind = (ns, name)

                if kind != None:
                    try:
                        data = self.convertType(string.join(self._data, ""),
                                                kind, attrs)
                    except UnknownTypeError:
                        data = None
                else:
                    data = None

                if data == None:
                    if self._data == None:
                        data = ''
                    else:
                        data = string.join(self._data, "")

                    if len(attrs) == 0:
                        try: data = str(data)
                        except: pass

                break

            data = structType(name = (ns, name), attrs = attrs)

            break

        if isinstance(data, compoundType):
            for i in range(len(cur)):
                v = cur.contents[i]
                data._addItem(cur.names[i], v, cur.subattrs[i])

                if isinstance(v, RefHolder):
                    v.parent = data

        if root:
            self._stack[-1].append(name, data, attrs)

        if idval != None:
            self._ids[idval] = data

            if self._refs.has_key(idval):
                for i in self._refs[idval]:
                    i.parent._placeItem(i.name, data, i.pos, i.subpos, attrs)

                del self._refs[idval]

        self.attrs[id(data)] = attrs

        if isinstance(data, anyType):
            data._setAttrs(attrs)

        self._data = None       # Stop accumulating

    def endDocument(self):
        if len(self._refs) == 1:
            raise Error, \
                "unresolved reference " + self._refs.keys()[0]
        elif len(self._refs) > 1:
            raise Error, \
                "unresolved references " + ', '.join(self._refs.keys())

    def startPrefixMapping(self, prefix, uri):
        self._prem[prefix] = uri
        self._prem_r[uri] = prefix

    def endPrefixMapping(self, prefix):
        try:
            del self._prem_r[self._prem[prefix]]
            del self._prem[prefix]
        except:
            pass

    def characters(self, c):
        if self._data != None:
            self._data.append(c)

    arrayre = '^(?:(?P<ns>[^:]*):)?' \
        '(?P<type>[^[]+)' \
        '(?:\[(?P<rank>,*)\])?' \
        '(?:\[(?P<asize>\d+(?:,\d+)*)?\])$'

    def startArray(self, name, kind, attrs, elemsname):
        if type(self.arrayre) == StringType:
            self.arrayre = re.compile (self.arrayre)

        offset = attrs.get((NS.ENC, "offset"))

        if offset != None:
            del attrs[(NS.ENC, "offset")]

            try:
                if offset[0] == '[' and offset[-1] == ']':
                    offset = int(offset[1:-1])
                    if offset < 0:
                        raise Exception
                else:
                    raise Exception
            except:
                raise AttributeError, "invalid Array offset"
        else:
            offset = 0

        try:
            m = self.arrayre.search(kind)

            if m == None:
                raise Exception

            t = m.group('type')

            if t == 'ur-type':
                return arrayType(None, name, attrs, offset, m.group('rank'),
                    m.group('asize'), elemsname)
            elif m.group('ns') != None:
                return typedArrayType(None, name,
                    (self._prem[m.group('ns')], t), attrs, offset,
                    m.group('rank'), m.group('asize'), elemsname)
            else:
                return typedArrayType(None, name, (None, t), attrs, offset,
                    m.group('rank'), m.group('asize'), elemsname)
        except:
            raise AttributeError, "invalid Array type `%s'" % kind

    # Conversion

    class DATETIMECONSTS:
        SIGNre = '(?P<sign>-?)'
        CENTURYre = '(?P<century>\d{2,})'
        YEARre = '(?P<year>\d{2})'
        MONTHre = '(?P<month>\d{2})'
        DAYre = '(?P<day>\d{2})'
        HOURre = '(?P<hour>\d{2})'
        MINUTEre = '(?P<minute>\d{2})'
        SECONDre = '(?P<second>\d{2}(?:\.\d*)?)'
        TIMEZONEre = '(?P<zulu>Z)|(?P<tzsign>[-+])(?P<tzhour>\d{2}):' \
            '(?P<tzminute>\d{2})'
        BOSre = '^\s*'
        EOSre = '\s*$'

        __allres = {'sign': SIGNre, 'century': CENTURYre, 'year': YEARre,
            'month': MONTHre, 'day': DAYre, 'hour': HOURre,
            'minute': MINUTEre, 'second': SECONDre, 'timezone': TIMEZONEre,
            'b': BOSre, 'e': EOSre}

        dateTime = '%(b)s%(sign)s%(century)s%(year)s-%(month)s-%(day)sT' \
            '%(hour)s:%(minute)s:%(second)s(%(timezone)s)?%(e)s' % __allres
        timeInstant = dateTime
        timePeriod = dateTime
        time = '%(b)s%(hour)s:%(minute)s:%(second)s(%(timezone)s)?%(e)s' % \
            __allres
        date = '%(b)s%(sign)s%(century)s%(year)s-%(month)s-%(day)s' \
            '(%(timezone)s)?%(e)s' % __allres
        century = '%(b)s%(sign)s%(century)s(%(timezone)s)?%(e)s' % __allres
        gYearMonth = '%(b)s%(sign)s%(century)s%(year)s-%(month)s' \
            '(%(timezone)s)?%(e)s' % __allres
        gYear = '%(b)s%(sign)s%(century)s%(year)s(%(timezone)s)?%(e)s' % \
            __allres
        year = gYear
        gMonthDay = '%(b)s--%(month)s-%(day)s(%(timezone)s)?%(e)s' % __allres
        recurringDate = gMonthDay
        gDay = '%(b)s---%(day)s(%(timezone)s)?%(e)s' % __allres
        recurringDay = gDay
        gMonth = '%(b)s--%(month)s--(%(timezone)s)?%(e)s' % __allres
        month = gMonth

        recurringInstant = '%(b)s%(sign)s(%(century)s|-)(%(year)s|-)-' \
            '(%(month)s|-)-(%(day)s|-)T' \
            '(%(hour)s|-):(%(minute)s|-):(%(second)s|-)' \
            '(%(timezone)s)?%(e)s' % __allres

        duration = '%(b)s%(sign)sP' \
            '((?P<year>\d+)Y)?' \
            '((?P<month>\d+)M)?' \
            '((?P<day>\d+)D)?' \
            '((?P<sep>T)' \
            '((?P<hour>\d+)H)?' \
            '((?P<minute>\d+)M)?' \
            '((?P<second>\d*(?:\.\d*)?)S)?)?%(e)s' % \
            __allres

        timeDuration = duration

        # The extra 31 on the front is:
        # - so the tuple is 1-based
        # - so months[month-1] is December's days if month is 1

        months = (31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

    def convertDateTime(self, value, kind):
        def getZoneOffset(d):
            zoffs = 0

            try:
                if d['zulu'] == None:
                    zoffs = 60 * int(d['tzhour']) + int(d['tzminute'])
                    if d['tzsign'] != '-':
                        zoffs = -zoffs
            except TypeError:
                pass

            return zoffs

        def applyZoneOffset(months, zoffs, date, minfield, posday = 1):
            if zoffs == 0 and (minfield > 4 or 0 <= date[5] < 60):
                return date

            if minfield > 5: date[5] = 0
            if minfield > 4: date[4] = 0

            if date[5] < 0:
                date[4] += int(date[5]) / 60
                date[5] %= 60

            date[4] += zoffs

            if minfield > 3 or 0 <= date[4] < 60: return date

            date[3] += date[4] / 60
            date[4] %= 60

            if minfield > 2 or 0 <= date[3] < 24: return date

            date[2] += date[3] / 24
            date[3] %= 24

            if minfield > 1:
                if posday and date[2] <= 0:
                    date[2] += 31       # zoffs is at most 99:59, so the
                                        # day will never be less than -3
                return date

            while 1:
                # The date[1] == 3 (instead of == 2) is because we're
                # going back a month, so we need to know if the previous
                # month is February, so we test if this month is March.

                leap = minfield == 0 and date[1] == 3 and \
                    date[0] % 4 == 0 and \
                    (date[0] % 100 != 0 or date[0] % 400 == 0)

                if 0 < date[2] <= months[date[1]] + leap: break

                date[2] += months[date[1] - 1] + leap

                date[1] -= 1

                if date[1] > 0: break

                date[1] = 12

                if minfield > 0: break

                date[0] -= 1

            return date

        try:
            exp = getattr(self.DATETIMECONSTS, kind)
        except AttributeError:
            return None

        if type(exp) == StringType:
            exp = re.compile(exp)
            setattr (self.DATETIMECONSTS, kind, exp)

        m = exp.search(value)

        try:
            if m == None:
                raise Exception

            d = m.groupdict()
            f = ('century', 'year', 'month', 'day',
                'hour', 'minute', 'second')
            fn = len(f)         # Index of first non-None value
            r = []

            if kind in ('duration', 'timeDuration'):
                if d['sep'] != None and d['hour'] == None and \
                    d['minute'] == None and d['second'] == None:
                    raise Exception

                f = f[1:]

                for i in range(len(f)):
                    s = d[f[i]]

                    if s != None:
                        if f[i] == 'second':
                            s = float(s)
                        else:
                            try: s = int(s)
                            except ValueError: s = long(s)

                        if i < fn: fn = i

                    r.append(s)

                if fn > len(r):         # Any non-Nones?
                    raise Exception

                if d['sign'] == '-':
                    r[fn] = -r[fn]

                return tuple(r)

            if kind == 'recurringInstant':
                for i in range(len(f)):
                    s = d[f[i]]

                    if s == None or s == '-':
                        if i > fn:
                            raise Exception
                        s = None
                    else:
                        if i < fn:
                            fn = i

                        if f[i] == 'second':
                            s = float(s)
                        else:
                            try:
                                s = int(s)
                            except ValueError:
                                s = long(s)

                    r.append(s)

                s = r.pop(0)

                if fn == 0:
                    r[0] += s * 100
                else:
                    fn -= 1

                if fn < len(r) and d['sign'] == '-':
                    r[fn] = -r[fn]

                cleanDate(r, fn)

                return tuple(applyZoneOffset(self.DATETIMECONSTS.months,
                    getZoneOffset(d), r, fn, 0))

            r = [0, 0, 1, 1, 0, 0, 0]

            for i in range(len(f)):
                field = f[i]

                s = d.get(field)

                if s != None:
                    if field == 'second':
                        s = float(s)
                    else:
                        try:
                            s = int(s)
                        except ValueError:
                            s = long(s)

                    if i < fn:
                        fn = i

                    r[i] = s

            if fn > len(r):     # Any non-Nones?
                raise Exception

            s = r.pop(0)

            if fn == 0:
                r[0] += s * 100
            else:
                fn -= 1

            if d.get('sign') == '-':
                r[fn] = -r[fn]

            cleanDate(r, fn)

            zoffs = getZoneOffset(d)

            if zoffs:
                r = applyZoneOffset(self.DATETIMECONSTS.months, zoffs, r, fn)

            if kind == 'century':
                return r[0] / 100

            s = []

            for i in range(1, len(f)):
                if d.has_key(f[i]):
                    s.append(r[i - 1])

            if len(s) == 1:
                return s[0]
            return tuple(s)
        except Exception, e:
            raise Error, "invalid %s value `%s' - %s" % (kind, value, e)

    intlimits = \
    {
        'nonPositiveInteger':   (0, None, 0),
        'non-positive-integer': (0, None, 0),
        'negativeInteger':      (0, None, -1),
        'negative-integer':     (0, None, -1),
        'long':                 (1, -9223372036854775808L,
                                    9223372036854775807L),
        'int':                  (0, -2147483648L, 2147483647),
        'short':                (0, -32768, 32767),
        'byte':                 (0, -128, 127),
        'nonNegativeInteger':   (0, 0, None),
        'non-negative-integer': (0, 0, None),
        'positiveInteger':      (0, 1, None),
        'positive-integer':     (0, 1, None),
        'unsignedLong':         (1, 0, 18446744073709551615L),
        'unsignedInt':          (0, 0, 4294967295L),
        'unsignedShort':        (0, 0, 65535),
        'unsignedByte':         (0, 0, 255),
    }
    floatlimits = \
    {
        'float':        (7.0064923216240861E-46, -3.4028234663852886E+38,
                         3.4028234663852886E+38),
        'double':       (2.4703282292062327E-324, -1.7976931348623158E+308,
                         1.7976931348623157E+308),
    }
    zerofloatre = '[1-9]'





    def convertType(self, d, t, attrs, config=Config):
        if t[0] is None and t[1] is not None:
            type = t[1].strip()
            if type[:9] == 'arrayType':
                index_eq = type.find('=')
                index_obr = type.find('[')
                index_cbr = type.find(']')
                elemtype = type[index_eq+1:index_obr]
                elemnum  = type[index_obr+1:index_cbr]
                if elemtype=="ur-type":
                    return(d)
                else:
                    newarr = map( lambda(di):
                                  self.convertToBasicTypes(d=di,
                                                       t = ( NS.XSD, elemtype),
                                                       attrs=attrs,
                                                       config=config),
                                  d)
                    return newarr
            else:
                t = (NS.XSD, t[1])

        return self.convertToBasicTypes(d, t, attrs, config)


    def convertToSOAPpyTypes(self, d, t, attrs, config=Config):
        pass


    def convertToBasicTypes(self, d, t, attrs, config=Config):
        dnn = d or ''

        #if Config.debug:
            #print "convertToBasicTypes:"
            #print "   requested_type=", t
            #print "   data=", d

        if t[0] in NS.EXSD_L:
            if t[1] == "integer":
                try:
                    d = int(d)
                    if len(attrs):
                        d = long(d)
                except:
                    d = long(d)
                return d
            if self.intlimits.has_key (t[1]): # integer types
                l = self.intlimits[t[1]]
                try: d = int(d)
                except: d = long(d)

                if l[1] != None and d < l[1]:
                    raise UnderflowError, "%s too small" % d
                if l[2] != None and d > l[2]:
                    raise OverflowError, "%s too large" % d

                if l[0] or len(attrs):
                    return long(d)
                return d
            if t[1] == "string":
                if len(attrs):
                    return unicode(dnn)
                try:
                    return str(dnn)
                except:
                    return dnn
            if t[1] == "boolean":
                d = d.strip().lower()
                if d in ('0', 'false'):
                    return 0
                if d in ('1', 'true'):
                    return 1
                raise AttributeError, "invalid boolean value"
            if t[1] in ('double','float'):
                l = self.floatlimits[t[1]]
                s = d.strip().lower()

                d = float(s)

                if config.strict_range:
                    if d < l[1]: raise UnderflowError
                    if d > l[2]: raise OverflowError
                else:
                    # some older SOAP impementations (notably SOAP4J,
                    # Apache SOAP) return "infinity" instead of "INF"
                    # so check the first 3 characters for a match.
                    if s == "nan":
                        return fpconst.NaN
                    elif s[0:3] in ("inf", "+inf"):
                        return fpconst.PosInf
                    elif s[0:3] == "-inf":
                        return fpconst.NegInf

                if fpconst.isNaN(d):
                    if s != 'nan':
                        raise ValueError, "invalid %s: %s" % (t[1], s)
                elif fpconst.isNegInf(d):
                    if s != '-inf':
                        raise UnderflowError, "%s too small: %s" % (t[1], s)
                elif fpconst.isPosInf(d):
                    if s != 'inf':
                        raise OverflowError, "%s too large: %s" % (t[1], s)
                elif d < 0 and d < l[1]:
                        raise UnderflowError, "%s too small: %s" % (t[1], s)
                elif d > 0 and ( d < l[0] or d > l[2] ):
                        raise OverflowError, "%s too large: %s" % (t[1], s)
                elif d == 0:
                    if type(self.zerofloatre) == StringType:
                        self.zerofloatre = re.compile(self.zerofloatre)

                    if self.zerofloatre.search(s):
                        raise UnderflowError, "invalid %s: %s" % (t[1], s)

                return d
            if t[1] in ("dateTime", "date", "timeInstant", "time"):
                return self.convertDateTime(d, t[1])
            if t[1] == "decimal":
                return float(d)
            if t[1] in ("language", "QName", "NOTATION", "NMTOKEN", "Name",
                "NCName", "ID", "IDREF", "ENTITY"):
                return collapseWhiteSpace(d)
            if t[1] in ("IDREFS", "ENTITIES", "NMTOKENS"):
                d = collapseWhiteSpace(d)
                return d.split()
        if t[0] in NS.XSD_L:
            if t[1] in ("base64", "base64Binary"):
                if d:
                    return base64.decodestring(d)
                else:
                    return ''
            if t[1] == "hexBinary":
                if d:
                    return decodeHexString(d)
                else:
                    return
            if t[1] == "anyURI":
                return urllib.unquote(collapseWhiteSpace(d))
            if t[1] in ("normalizedString", "token"):
                return collapseWhiteSpace(d)
        if t[0] == NS.ENC:
            if t[1] == "base64":
                if d:
                    return base64.decodestring(d)
                else:
                    return ''
        if t[0] == NS.XSD:
            if t[1] == "binary":
                try:
                    e = attrs[(None, 'encoding')]

                    if d:
                        if e == 'hex':
                            return decodeHexString(d)
                        elif e == 'base64':
                            return base64.decodestring(d)
                    else:
                        return ''
                except:
                    pass

                raise Error, "unknown or missing binary encoding"
            if t[1] == "uri":
                return urllib.unquote(collapseWhiteSpace(d))
            if t[1] == "recurringInstant":
                return self.convertDateTime(d, t[1])
        if t[0] in (NS.XSD2, NS.ENC):
            if t[1] == "uriReference":
                return urllib.unquote(collapseWhiteSpace(d))
            if t[1] == "timePeriod":
                return self.convertDateTime(d, t[1])
            if t[1] in ("century", "year"):
                return self.convertDateTime(d, t[1])
        if t[0] in (NS.XSD, NS.XSD2, NS.ENC):
            if t[1] == "timeDuration":
                return self.convertDateTime(d, t[1])
        if t[0] == NS.XSD3:
            if t[1] == "anyURI":
                return urllib.unquote(collapseWhiteSpace(d))
            if t[1] in ("gYearMonth", "gMonthDay"):
                return self.convertDateTime(d, t[1])
            if t[1] == "gYear":
                return self.convertDateTime(d, t[1])
            if t[1] == "gMonth":
                return self.convertDateTime(d, t[1])
            if t[1] == "gDay":
                return self.convertDateTime(d, t[1])
            if t[1] == "duration":
                return self.convertDateTime(d, t[1])
        if t[0] in (NS.XSD2, NS.XSD3):
            if t[1] == "token":
                return collapseWhiteSpace(d)
            if t[1] == "recurringDate":
                return self.convertDateTime(d, t[1])
            if t[1] == "month":
                return self.convertDateTime(d, t[1])
            if t[1] == "recurringDay":
                return self.convertDateTime(d, t[1])
        if t[0] == NS.XSD2:
            if t[1] == "CDATA":
                return collapseWhiteSpace(d)

        raise UnknownTypeError, "unknown type `%s'" % (str(t[0]) + ':' + t[1])


################################################################################
# call to SOAPParser that keeps all of the info
################################################################################
def _parseSOAP(xml_str, rules = None):
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO

    parser = xml.sax.make_parser()
    t = SOAPParser(rules = rules)
    parser.setContentHandler(t)
    e = xml.sax.handler.ErrorHandler()
    parser.setErrorHandler(e)

    inpsrc = xml.sax.xmlreader.InputSource()
    inpsrc.setByteStream(StringIO(xml_str))

    # turn on namespace mangeling
    parser.setFeature(xml.sax.handler.feature_namespaces,1)

    try:
        parser.parse(inpsrc)
    except xml.sax.SAXParseException, e:
        parser._parser = None
        raise e
    
    return t

################################################################################
# SOAPParser's more public interface
################################################################################
def parseSOAP(xml_str, attrs = 0):
    t = _parseSOAP(xml_str)

    if attrs:
        return t.body, t.attrs
    return t.body


def parseSOAPRPC(xml_str, header = 0, body = 0, attrs = 0, rules = None):

    t = _parseSOAP(xml_str, rules = rules)
    p = t.body[0]

    # Empty string, for RPC this translates into a void
    if type(p) in (type(''), type(u'')) and p in ('', u''):
        name = "Response"
        for k in t.body.__dict__.keys():
            if k[0] != "_":
                name = k
        p = structType(name)
        
    if header or body or attrs:
        ret = (p,)
        if header : ret += (t.header,)
        if body: ret += (t.body,)
        if attrs: ret += (t.attrs,)
        return ret
    else:
        return p



#!/usr/bin/python
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for accessing all other services."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import os
import thread
import time

from adspygoogle.adwords import AdWordsSanityCheck
from adspygoogle.adwords import AUTH_TOKEN_SERVICE
from adspygoogle.adwords import DEFAULT_API_VERSION
from adspygoogle.adwords import LIB_SIG
from adspygoogle.adwords import REQUIRED_SOAP_HEADERS
from adspygoogle.adwords.GenericAdWordsService import GenericAdWordsService
from adspygoogle.adwords.ReportDownloader import ReportDownloader
from adspygoogle.common import SanityCheck
from adspygoogle.common import Utils
from adspygoogle.common.Client import Client
from adspygoogle.common.Errors import AuthTokenError
from adspygoogle.common.Errors import ValidationError
from adspygoogle.common.Logger import Logger


class AdWordsClient(Client):

  """Provides entry point to all web services.

  Allows instantiation of all AdWords API web services.
  """

  auth_pkl_name = 'adwords_api_auth.pkl'
  config_pkl_name = 'adwords_api_config.pkl'

  def __init__(self, headers=None, config=None, path=None,
               login_token=None, login_captcha=None):
    """Inits AdWordsClient.

    Args:
      [optional]
      headers: dict Object with populated authentication credentials.
      config: dict Object with client configuration values.
      path: str Relative or absolute path to home directory (i.e. location of
            pickles and logs/).
      login_token: str Token representing the specific CAPTCHA challenge.
      login_captcha: str String entered by the user as an answer to a CAPTCHA
                     challenge.

    Example:
      headers = {
        'email': 'johndoe@example.com',
        'password': 'secret',
        'authToken': '...',
        'clientCustomerId': '1234567890',
        'userAgent': 'GoogleTest',
        'developerToken': 'ABcdeFGH93KL-NOPQ_STUv.',
        'validateOnly': 'n',
        'partialFailure': 'n',
        'oauth2credentials': 'See use_oauth2.py'.
      }
      config = {
        'home': '/path/to/home',
        'log_home': '/path/to/logs/home',
        'proxy': 'http://example.com:8080',
        'xml_parser': '1', # PYXML = 1, ELEMENTREE = 2
        'debug': 'n',
        'raw_debug': 'n',
        'xml_log': 'y',
        'request_log': 'y',
        'raw_response': 'n',
        'strict': 'y',
        'pretty_xml': 'y',
        'compress': 'y',
        'access': ''
      }
      path = '/path/to/home'
    """
    super(AdWordsClient, self).__init__(headers, config, path)

    self.__lock = thread.allocate_lock()
    self.__loc = None

    if path is not None:
      # Update absolute path for a given instance of AdWordsClient, based on
      # provided relative path.
      if os.path.isabs(path):
        AdWordsClient.home = path
      else:
        # NOTE(api.sgrinberg): Keep first parameter of join() as os.getcwd(),
        # do not change it to AdWordsClient.home. Otherwise, may break when
        # multiple instances of AdWordsClient exist during program run.
        AdWordsClient.home = os.path.join(os.getcwd(), path)

      # If pickles don't exist at given location, default to "~".
      if (not headers and not config and
          (not os.path.exists(os.path.join(AdWordsClient.home,
                                           AdWordsClient.auth_pkl_name)) or
           not os.path.exists(os.path.join(AdWordsClient.home,
                                           AdWordsClient.config_pkl_name)))):
        AdWordsClient.home = os.path.expanduser('~')
    else:
      AdWordsClient.home = os.path.expanduser('~')

    # Update location for both pickles.
    AdWordsClient.auth_pkl = os.path.join(AdWordsClient.home,
                                          AdWordsClient.auth_pkl_name)
    AdWordsClient.config_pkl = os.path.join(AdWordsClient.home,
                                            AdWordsClient.config_pkl_name)

    # Only load from the pickle if config wasn't specified.
    self._config = config or self.__LoadConfigValues()
    self._config = self.__SetMissingDefaultConfigValues(self._config)
    self._config['home'] = AdWordsClient.home

    # Validate XML parser to use.
    SanityCheck.ValidateConfigXmlParser(self._config['xml_parser'])

    # Initialize units and operations for current instance of AdWordsClient
    # object (using list to take advantage of Python's pass-by-reference).
    self._config['units'] = [0]
    self._config['operations'] = [0]
    self._config['last_units'] = [0]
    self._config['last_operations'] = [0]

    # Only load from the pickle if 'headers' wasn't specified.
    if headers is None:
      self._headers = self.__LoadAuthCredentials()
    else:
      if Utils.BoolTypeConvert(self._config['strict']):
        SanityCheck.ValidateRequiredHeaders(headers, REQUIRED_SOAP_HEADERS)
      self._headers = headers

    # Internally, store user agent as 'userAgent'.
    if 'useragent' in self._headers:
      self._headers['userAgent'] = self._headers['useragent']
      self._headers = Utils.UnLoadDictKeys(self._headers, ['useragent'])
    if Utils.BoolTypeConvert(self._config['strict']):
      SanityCheck.ValidateRequiredHeaders(self._headers,
                                          REQUIRED_SOAP_HEADERS)

    # Load validateOnly header, if one was set.
    if 'validateOnly' in self._headers:
      self._headers['validateOnly'] = str(Utils.BoolTypeConvert(
          self._headers['validateOnly'])).lower()

    # Load partialFailure header, if one was set.
    if 'partialFailure' in self._headers:
      self._headers['partialFailure'] = str(Utils.BoolTypeConvert(
          self._headers['partialFailure'])).lower()

    # Load/set authentication token.
    if self._headers.get('authToken'):
      # If they have a non-empty authToken, set the epoch and skip the rest.
      self._config['auth_token_epoch'] = time.time()
    elif self._headers.get('oauth2credentials'):
      # If they have oauth2credentials, that's also fine.
      pass
    elif (self._headers.get('email') and self._headers.get('password')
          and not self._headers.get('authToken')):
      # If they have a non-empty email and password but no or empty authToken,
      # generate an authToken.
      try:
        self._headers['authToken'] = Utils.GetAuthToken(
            self._headers['email'], self._headers['password'],
            AUTH_TOKEN_SERVICE, LIB_SIG, self._config['proxy'], login_token,
            login_captcha)
        self._config['auth_token_epoch'] = time.time()
      except AuthTokenError, e:
        # We would end up here if non-valid Google Account's credentials were
        # specified.
        raise ValidationError('Was not able to obtain an AuthToken for '
                              'provided email and password, see root_cause.', e)
    else:
      # We need either oauth2credentials OR authToken.
      raise ValidationError('Authentication data is missing.')

    self.__is_mcc = False

    # Initialize logger.
    self.__logger = Logger(LIB_SIG, self._config['log_home'])

  def __LoadAuthCredentials(self):
    """Load existing authentication credentials from adwords_api_auth.pkl.

    Returns:
      dict Dictionary object with populated authentication credentials.
    """
    return super(AdWordsClient, self)._LoadAuthCredentials()

  def __WriteUpdatedAuthValue(self, key, new_value):
    """Write updated authentication value for a key in adwords_api_auth.pkl.

    Args:
      key: str Key to update.
      new_value: str New value to update the key with.
    """
    super(AdWordsClient, self)._WriteUpdatedAuthValue(key, new_value)

  def __LoadConfigValues(self):
    """Load existing configuration values from adwords_api_config.pkl.

    Returns:
      dict Dictionary object with populated configuration values.
    """
    return super(AdWordsClient, self)._LoadConfigValues()

  def __SetMissingDefaultConfigValues(self, config=None):
    """Set default configuration values for missing elements in the config dict.

    Args:
      config: dict Object with client configuration values.

    Returns:
      dict A config dictionary with default values set.
    """
    if config is None: config = {}
    config = super(AdWordsClient, self)._SetMissingDefaultConfigValues(config)
    default_config = {
        'home': AdWordsClient.home,
        'log_home': os.path.join(AdWordsClient.home, 'logs')
    }
    for key in default_config:
      if key not in config:
        config[key] = default_config[key]
    return config

  def GetUnits(self):
    """Return number of API units consumed by current instance of AdWordsClient
    object.

    Returns:
      int Number of API units.
    """
    return self._config['units'][0]

  def GetOperations(self):
    """Return number of API ops performed by current instance of AdWordsClient
    object.

    Returns:
      int Number of API operations.
    """
    return self._config['operations'][0]

  def GetLastUnits(self):
    """Return number of API units consumed by last API call.

    Returns:
      int Number of API units.
    """
    return self._config['last_units'][0]

  def GetLastOperations(self):
    """Return number of API ops performed by last API call.

    Returns:
      int Number of API operations.
    """
    return self._config['last_operations'][0]

  def UseMcc(self, state):
    """Choose to make an API request against MCC account or a sub-account.

    Args:
      state: bool State of the API request, whether to use MCC.
    """
    self.__is_mcc = False
    if state:
      self.__is_mcc = True

  def __GetUseMcc(self):
    """Return current state of the API request.

    Returns:
      bool State of the API request, whether to use MCC.
    """
    return self.__is_mcc

  def __SetUseMcc(self, state):
    """Chooses to make an API request against MCC account or a sub-account.

    Args:
      state: bool State of the API request, whether to use MCC.
    """
    self.__is_mcc = state

  use_mcc = property(__GetUseMcc, __SetUseMcc)

  def SetClientCustomerId(self, client_customer_id):
    """Temporarily change client customer id for a given AdWordsClient instance.

    Args:
      client_customer_id: str New client customer id to use.
    """
    if ('clientCustomerId' not in self._headers or
        self._headers['clientCustomerId'] != client_customer_id):
      self._headers['clientCustomerId'] = client_customer_id

  def __GetValidateOnly(self):
    """Return current state of the validation mode.

    Returns:
      bool State of the validation mode.
    """
    return self._headers['validateOnly']

  def __SetValidateOnly(self, value):
    """Temporarily change validation mode for a given AdWordsClient instance.

    Args:
      value: mixed New state of the validation mode using BoolTypeConvert.
    """
    self._headers['validateOnly'] = str(Utils.BoolTypeConvert(value)).lower()

  validate_only = property(__GetValidateOnly, __SetValidateOnly)

  def __GetPartialFailure(self):
    """Return current state of the partial failure mode.

    Returns:
      bool State of the partial failure mode.
    """
    return self._headers['partialFailure']

  def __SetPartialFailure(self, value):
    """Temporarily change partial failure mode for a given AdWordsClient
    instance.

    Args:
      value: mixed New state of the partial failure mode using BoolTypeConvert.
    """
    self._headers['partialFailure'] = str(Utils.BoolTypeConvert(value)).lower()

  partial_failure = property(__GetPartialFailure, __SetPartialFailure)

  def __GetAuthCredentialsForAccessLevel(self):
    """Return auth credentials based on the access level of the request.

    Request can have an MCC level access or a sub account level access.

    Returns:
      dict Authentiaction credentials.
    """
    old_headers = self.GetAuthCredentials()
    new_headers = {}
    is_mcc = self.__is_mcc

    for key, value in old_headers.iteritems():
      new_headers[key] = value
      if key == 'clientCustomerId':
        if is_mcc and 'email' in old_headers:
          new_headers[key] = None

    return new_headers

  def CallRawMethod(self, soap_message, url, server, http_proxy):
    """Call API method directly, using raw SOAP message.

    For API calls performed with this method, outgoing data is not run through
    library's validation logic.

    Args:
      soap_message: str SOAP XML message.
      url: str URL of the API service for the method to call.
      server: str API server to access for this API call.
      http_proxy: str HTTP proxy to use for this API call.

    Returns:
      tuple Response from the API method (SOAP XML response message).
    """
    service_name = url.split('/')[-1]
    service = getattr(self, 'Get' + service_name)(server=server,
                                                  http_proxy=http_proxy)
    return service.CallRawMethod(soap_message)

  def GetAdExtensionOverrideService(self, server='https://adwords.google.com',
                                    version=None, http_proxy=None):
    """Call API method in AdExtensionOverrideService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdExtensionOverrideService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)
    AdWordsSanityCheck.ValidateService('AdExtensionOverrideService', version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdExtensionOverrideService')

  def GetAdGroupAdService(self, server='https://adwords.google.com',
                          version=None, http_proxy=None):
    """Call API method in AdGroupAdService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdGroupAdService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdGroupAdService')

  def GetAdGroupBidModifierService(self, server='https://adwords.google.com',
                                   version=None, http_proxy=None):
    """Call API method in AdGroupBidModifierService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdGroupBidModifierService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdGroupBidModifierService')

  def GetAdGroupCriterionService(self, server='https://adwords.google.com',
                                 version=None, http_proxy=None):
    """Call API method in AdGroupCriterionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdGroupCriterionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdGroupCriterionService')

  def GetAdGroupFeedService(self, server='https://adwords.google.com',
                            version=None, http_proxy=None):
    """Creates an object used to call methods in the AdGroupFeedService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdGroupFeedService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdGroupFeedService')

  def GetAdGroupService(self, server='https://adwords.google.com',
                        version=None, http_proxy=None):
    """Call API method in AdGroupService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdGroupService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdGroupService')

  def GetAdParamService(self, server='https://adwords.google.com',
                        version=None, http_proxy=None):
    """Call API method in AdParamService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdParamService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdParamService')

  def GetAdwordsUserListService(self, server='https://adwords.google.com',
                                version=None, http_proxy=None):
    """Call API method in AdwordsUserListService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AdwordsUserListService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'rm',
        'default_group': 'rm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AdwordsUserListService')

  def GetAlertService(self, server='https://adwords.google.com', version=None,
                      http_proxy=None):
    """Call API method in AlertService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of AlertService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'mcm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'AlertService')

  def GetBiddingStrategyService(self, server='https://adwords.google.com',
                                version=None, http_proxy=None):
    """Call API method in BiddingStrategyService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of BiddingStrategyService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'BiddingStrategyService')

  def GetBudgetService(self, server='https://adwords.google.com',
                       version=None, http_proxy=None):
    """Call API method in BudgetOrderService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of BudgetService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'BudgetService')

  def GetBudgetOrderService(self, server='https://adwords.google.com',
                            version=None, http_proxy=None):
    """Call API method in BudgetOrderService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of BudgetOrderService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'billing',
        'default_group': 'billing',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'BudgetOrderService')

  def GetMutateJobService(self, server='https://adwords.google.com',
                              version=None, http_proxy=None):
    """Call API method in MutateJobService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of BulkMutateJobService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'MutateJobService')

  def GetCampaignAdExtensionService(self, server='https://adwords.google.com',
                                    version=None, http_proxy=None):
    """Call API method in CampaignAdExtensionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CampaignAdExtensionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CampaignAdExtensionService')

  def GetCampaignCriterionService(self, server='https://adwords.google.com',
                                  version=None, http_proxy=None):
    """Call API method in CampaignCriterionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CampaignCriterionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CampaignCriterionService')

  def GetCampaignFeedService(self, server='https://adwords.google.com',
                             version=None, http_proxy=None):
    """Creates an object used to call methods in the CampaignFeedService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CampaignFeedService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CampaignFeedService')

  def GetCampaignSharedSetService(self, server='https://adwords.google.com',
                                  version=None, http_proxy=None):
    """Creates an object used to call methods in the CampaignSharedSetService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CampaignSharedSetService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CampaignSharedSetService')

  def GetCampaignService(self, server='https://adwords.google.com',
                         version=None, http_proxy=None):
    """Call API method in CampaignService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CampaignService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CampaignService')

  def GetConstantDataService(self, server='https://adwords.google.com',
                             version=None, http_proxy=None):
    """Call API method in ConstantDataService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of ConstantDataService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'ConstantDataService')

  def GetCustomerService(self, server='https://adwords.google.com',
                         version=None, http_proxy=None):
    """Call API method in CustomerService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CustomerService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'mcm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CustomerService')

  def GetCustomerSyncService(self, server='https://adwords.google.com',
                             version=None, http_proxy=None):
    """Call API method in CustomerSyncService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of CustomerSyncService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'ch',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'CustomerSyncService')

  def GetExperimentService(self, server='https://adwords.google.com',
                           version=None, http_proxy=None):
    """Call API method in ExperimentService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of ExperimentService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'ExperimentService')

  def GetFeedItemService(self, server='https://adwords.google.com',
                         version=None, http_proxy=None):
    """Creates an object used to call methods in the FeedItemService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of FeedItemService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'FeedItemService')

  def GetFeedMappingService(self, server='https://adwords.google.com',
                            version=None, http_proxy=None):
    """Creates an object used to call methods in the FeedMappingService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of FeedMappingService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'FeedMappingService')

  def GetFeedService(self, server='https://adwords.google.com',
                     version=None, http_proxy=None):
    """Creates an object used to call methods in the FeedService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of FeedService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'FeedService')

  def GetGeoLocationService(self, server='https://adwords.google.com',
                            version=None, http_proxy=None):
    """Call API method in GeoLocationService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of GeoLocationService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'GeoLocationService')

  def GetLocationCriterionService(self, server='https://adwords.google.com',
                                  version=None, http_proxy=None):
    """Call API method in LocationCriterionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      LocationCriterionService New instance of LocationCriterionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'LocationCriterionService')

  def GetManagedCustomerService(self, server='https://adwords.google.com',
                                version=None, http_proxy=None):
    """Call API method in ManagedCustomerService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of ManagedCustomerService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'mcm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'ManagedCustomerService')

  def GetMediaService(self, server='https://adwords.google.com', version=None,
                      http_proxy=None):
    """Call API method in MediaService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of MediaService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'MediaService')

  def GetOfflineConversionFeedService(self, server='https://adwords.google.com',
                                      version=None, http_proxy=None):
    """Call API method in OfflineConversionFeedServiice.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of OfflineConversionFeedService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'OfflineConversionFeedService')

  def GetReportDefinitionService(self, server='https://adwords.google.com',
                                 version=None, http_proxy=None):
    """Call API method in ReportDefinitionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of ReportDefinitionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'ReportDefinitionService')

  def GetReportDownloader(self, server='https://adwords.google.com',
                          version=None, http_proxy=None):
    """Returns an instance of ReportDownloader, used to download reports.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      ReportService New instance of ReportDownloader object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return ReportDownloader(headers, self._config, op_config, self.__logger)

  def GetSharedCriterionService(self, server='https://adwords.google.com',
                                version=None, http_proxy=None):
    """Creates an object used to call methods in the SharedCriterionService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of SharedCriterionService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'SharedCriterionService')

  def GetSharedSetService(self, server='https://adwords.google.com',
                          version=None, http_proxy=None):
    """Creates an object used to call methods in the SharedSetService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of SharedSetService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'SharedSetService')

  def GetTargetingIdeaService(self, server='https://adwords.google.com',
                              version=None, http_proxy=None):
    """Call API method in TargetingIdeaService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of TargetingIdeaService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'o',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'TargetingIdeaService')

  def GetTrafficEstimatorService(self, server='https://adwords.google.com',
                                 version=None, http_proxy=None):
    """Call API method in TrafficEstimatorService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of TrafficEstimatorService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'o',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'TrafficEstimatorService')

  def GetUserListService(self, server='https://adwords.google.com',
                         version=None, http_proxy=None):
    """Call API method in UserListService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of UserListService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)
    AdWordsSanityCheck.ValidateService('UserListService', version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'UserListService')

  def GetConversionTrackerService(self, server='https://adwords.google.com',
                                  version=None, http_proxy=None):
    """Call API method in ConversionTrackerService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of ConversionTrackerService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'ConversionTrackerService')

  def GetDataService(self, server='https://adwords.google.com',
                     version=None, http_proxy=None):
    """Call API method in DataService.

    Args:
      [optional]
      server: str API server to access for this API call. The only supported
              server is currently 'https://adwords.google.com'.
      version: str API version to use.
      http_proxy: str HTTP proxy to use.

    Returns:
      GenericAdWordsService New instance of DataService object.
    """
    headers = self.__GetAuthCredentialsForAccessLevel()

    if version is None:
      version = DEFAULT_API_VERSION
    if Utils.BoolTypeConvert(self._config['strict']):
      AdWordsSanityCheck.ValidateServer(server, version)

    # Load additional configuration data.
    op_config = {
        'server': server,
        'version': version,
        'group': 'cm',
        'default_group': 'cm',
        'http_proxy': http_proxy
    }
    return GenericAdWordsService(headers, self._config, op_config, self.__lock,
                                 self.__logger, 'DataService')



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Settings and configurations for the client library."""

__author__ = 'api.sgrinberg@gmail.com (Stan Grinberg)'

import sys

VERSION = '3.1.1'

MIN_PY_VERSION = '2.4.4'
PYXML_NAME = 'PyXML'
ETREE_NAME = 'ElementTree'
MIN_PYXML_VERSION = '0.8.3'
MIN_ETREE_VERSION = '1.2.6'
PYXML = '1'
ETREE = '2'
COMMON_LIB_SIG = 'Common-Python/%s' % VERSION
PYTHON_VERSION = 'Python/%d.%d' % (sys.version_info[0], sys.version_info[1])


def GenerateLibSig(short_name, version):
  """Generates a library signature suitable for a UserAgent field.

  Args:
    short_name: str The short, product-specific name of the library.
    version: str The product-specific version of the library.
  Returns:
    str Library signature suitable to append to user-supplied user-agent value.
  """
  return ' (%s/%s, %s, %s)' % (short_name, version, COMMON_LIB_SIG,
                               PYTHON_VERSION)



"""Translate strings to and from SOAP 1.2 XML name encoding

Implements rules for mapping application defined name to XML names
specified by the w3 SOAP working group for SOAP version 1.2 in
Appendix A of "SOAP Version 1.2 Part 2: Adjuncts", W3C Working Draft
17, December 2001, <http://www.w3.org/TR/soap12-part2/#namemap>

Also see <http://www.w3.org/2000/xp/Group/xmlp-issues>.

Author: Gregory R. Warnes <Gregory.R.Warnes@Pfizer.com>
Date::  2002-04-25
Version 0.9.0

"""

ident = "$Id: XMLname.py,v 1.4 2005/02/16 14:45:37 warnes Exp $"

from re import *


def _NCNameChar(x):
    return x.isalpha() or x.isdigit() or x=="." or x=='-' or x=="_" 


def _NCNameStartChar(x):
    return x.isalpha() or x=="_" 


def _toUnicodeHex(x):
    hexval = hex(ord(x[0]))[2:]
    hexlen = len(hexval)
    # Make hexval have either 4 or 8 digits by prepending 0's
    if   (hexlen==1): hexval = "000" + hexval
    elif (hexlen==2): hexval = "00"  + hexval
    elif (hexlen==3): hexval = "0"   + hexval
    elif (hexlen==4): hexval = ""    + hexval
    elif (hexlen==5): hexval = "000" + hexval
    elif (hexlen==6): hexval = "00"  + hexval
    elif (hexlen==7): hexval = "0"   + hexval
    elif (hexlen==8): hexval = ""    + hexval    
    else: raise Exception, "Illegal Value returned from hex(ord(x))"
    
    return "_x"+ hexval + "_"


def _fromUnicodeHex(x):
    return eval( r'u"\u'+x[2:-1]+'"' ) 


def toXMLname(string):
    """Convert string to a XML name."""
    if string.find(':') != -1 :
        (prefix, localname) = string.split(':',1)
    else:
        prefix = None
        localname = string
    
    T = unicode(localname)

    N = len(localname)
    X = [];
    for i in range(N) :
        if i< N-1 and T[i]==u'_' and T[i+1]==u'x':
            X.append(u'_x005F_')
        elif i==0 and N >= 3 and \
                 ( T[0]==u'x' or T[0]==u'X' ) and \
                 ( T[1]==u'm' or T[1]==u'M' ) and \
                 ( T[2]==u'l' or T[2]==u'L' ):
            X.append(u'_xFFFF_' + T[0])
        elif (not _NCNameChar(T[i])) or (i==0 and not _NCNameStartChar(T[i])):
            X.append(_toUnicodeHex(T[i]))
        else:
            X.append(T[i])
    
    if prefix:
        return "%s:%s" % (prefix, u''.join(X))
    return u''.join(X)


def fromXMLname(string):
    """Convert XML name to unicode string."""

    retval = sub(r'_xFFFF_','', string )

    def fun( matchobj ):
        return _fromUnicodeHex( matchobj.group(0) )

    retval = sub(r'_x[0-9A-Za-z]+_', fun, retval )
        
    return retval
